mod button;
use std::cell::Cell;

pub use button::ButtonSet;
use kahlo::math::PixelBox;
use winit::keyboard::Key;

use crate::layout::LayoutNodeID;

#[derive(Clone, Copy, Debug)]
#[repr(u8)]
pub enum MouseButton {
    Left,
    Right,
    Middle,
    ScrollUp,
    ScrollDown,
    ExtraButton1,
    ExtraButton2,
    ExtraButton3,
    ExtraButton4,
}

impl From<MouseButton> for u8 {
    fn from(mb: MouseButton) -> u8 {
        mb as u8
    }
}

pub type MouseButtonSet = ButtonSet<MouseButton>;

#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
pub enum MouseCheckStatus {
    /// Cursor is outside the region.
    Idle,
    /// Cursor just entered the region.
    Enter,
    /// Cursor just left the region.
    Leave,

    /// Cursor is still in the region, no mouse button held.
    Hover,
    /// Cursor is still in the region, with the mouse button pressed.
    Hold,

    /// Cursor is in the region, and the button was just pressed.
    Click,
    /// Cursor is in the region, and the button was just released.
    Release,
}

impl MouseCheckStatus {
    /// is the cursor outside the region?
    pub fn is_idle(&self) -> bool {
        *self == Self::Idle
    }

    /// did the cursor just enter the region?
    pub fn is_enter(&self) -> bool {
        *self == Self::Enter
    }

    /// did the cursor just leave the region?
    pub fn is_leave(&self) -> bool {
        *self == Self::Leave
    }

    /// is the cursor inside the region with no mouse buttons active?
    pub fn is_hover(&self) -> bool {
        *self == Self::Hover
    }

    /// is the cursor inside the region with a mouse button held?
    pub fn is_hold(&self) -> bool {
        *self == Self::Hold
    }

    /// is the cursor inside the region and the mouse button just pressed?
    pub fn is_click(&self) -> bool {
        *self == Self::Click
    }

    /// is the cursor inside the region and the mouse button just released?
    pub fn is_release(&self) -> bool {
        *self == Self::Release
    }
}

#[derive(Default)]
pub struct MouseState {
    pub pos: kahlo::math::PixelPoint,
    pub last_pos: kahlo::math::PixelPoint,
    pub buttons: ButtonSet<MouseButton>,
    pub last_buttons: ButtonSet<MouseButton>,
}

impl MouseState {
    pub fn pressed(&self) -> MouseButtonSet {
        self.buttons & !self.last_buttons
    }
    pub fn released(&self) -> MouseButtonSet {
        !self.buttons & self.last_buttons
    }
}

#[derive(Default)]
pub struct InputState {
    pub mouse: MouseState,
    pub text_input: Option<String>,
    pub keypress: Option<Key>,
    focus: Cell<Option<LayoutNodeID>>,
}

impl InputState {
    pub fn mouse_status_in(&self, area: PixelBox, button: MouseButton) -> MouseCheckStatus {
        match (
            area.contains(self.mouse.pos),
            area.contains(self.mouse.last_pos),
        ) {
            (false, false) => MouseCheckStatus::Idle,
            (true, false) => MouseCheckStatus::Enter,
            (false, true) => MouseCheckStatus::Leave,
            (true, true) => match (
                self.mouse.buttons.active(button),
                self.mouse.last_buttons.active(button),
            ) {
                (false, false) => MouseCheckStatus::Hover,
                (true, false) => MouseCheckStatus::Click,
                (false, true) => MouseCheckStatus::Release,
                (true, true) => MouseCheckStatus::Hold,
            },
        }
    }
}

/// Focus-related functionality
impl InputState {
    pub fn focus_on(&self, id: LayoutNodeID) {
        self.focus.set(Some(id));
    }
    pub fn clear_focus(&self) {
        self.focus.set(None);
    }
    pub fn is_focused(&self, id: LayoutNodeID) -> bool {
        self.focus.get() == Some(id)
    }
}

/// Internal updating functions
impl InputState {
    /// Note the passing of a frame, and clear per-frame input information.
    pub(crate) fn tick(&mut self) {
        self.mouse.last_pos = self.mouse.pos;
        self.mouse.last_buttons = self.mouse.buttons;
        self.text_input.take();
        self.keypress.take();
    }
    pub(crate) fn set_text_input(&mut self, input: String) {
        self.text_input = Some(input);
    }
}
