//! Layout calculations for the widget tree. All layout calculations are performed on a tree of
//! [`LayoutNode`]s, which are stored in various widget structs.
//!
//! Layout calculations are done in two passes:
//! - **Arrangement**: computes minimum sizes, net size policies, and records parent/child relationships between
//! nodes for optimization purposes.
//! - **Layout**: given minimum sizes and net size policies from the arrangement step, computes
//! exact pixel values for positions and sizes by distributing the available space appropriately.
//!
//! Collectively, layout nodes form a tree structure with _externally_-stored edges. This is a
//! slightly odd data structure, so it deserves some elaboration. In order to play nicely with
//! Rust's borrow checker without using interior mutability (which comes with a large set of
//! gotchas and a performance hit), when using internally-stored edges, layout nodes must be
//! traversed from the root node to acquire a mutable reference to an interior node.
//!
//! To make this less awkward, the tree structure of the layout nodes -- and indeed the ownership
//! of nodes -- is not stored directly in layout nodes themselves, but instead in a separate widget
//! tree structure. At first this seems an unnecessary complication and major source of
//! awkwardness, but it allows for a few very useful tricks:
//! - Mutable references can be created for any number of nodes,
//! - The entire layout tree need not be traversed to find a specific child,
//! - There is no 1:1 requirement between widgets and layout nodes, so layout nodes can be used for
//! internal alignment,
//! - Layout is significantly easier to debug as it is by design entirely agnostic to details of
//! the widgets involved,
//! - Rather than numerical indices or similar awkward indexing schemes, nodes are simply stored as
//! members of a struct and benefit from all the usual borrow checker behaviour.
//!
//! The tree structure is created by implementations of the [`LayoutTreeNode`] trait. Consider the
//! following tree:
//! ```text
//!     root --+-- child1
//!            |
//!            \-- child2 --+-- child2a
//!                         |
//!                         \-- child2b
//! ```
//!
//! The following is one simplified way the structure can be imposed externally:
//! ```
//! # use patina::layout::*;
//! trait Widget {
//!     fn lnode(&self) -> LayoutNodeAccess<'_>;
//! }
//! struct RootWidget {
//!     node: LayoutNode,
//!     label: LabelWidget,
//!     select: ChoiceWidget,
//! }
//! impl LayoutTreeNode<()> for RootWidget {
//!     fn current_node(&self) -> &LayoutNode { &self.node }
//!     fn child_count(&self) -> usize { 2 }
//!     fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'_>> {
//!         match ndx {
//!             0 => Some(self.label.lnode()),  // root -> child1 edge
//!             1 => Some(self.select.lnode()), // root -> child2 edge
//!             _ => None
//!         }
//!     }
//! }
//! impl Widget for RootWidget {
//!     fn lnode(&self) -> LayoutNodeAccess<'_> {
//!         LayoutNodeAccess::new::<()>(self)
//!     }
//! }
//! struct LabelWidget {
//!     // see documentation on LeafLayoutNode for more details
//!     // this is basically a shortcut to avoid reimplementing LayoutTreeNode
//!     // to store a node with no children
//!     node: LeafLayoutNode
//! }
//! impl Widget for LabelWidget {
//!     fn lnode(&self) -> LayoutNodeAccess<'_> {
//!         self.node.access()
//!     }
//! }
//! struct ChoiceWidget {
//!     node: LayoutNode,
//!     choice1: LabelWidget,
//!     choice2: LabelWidget,
//! }
//! impl LayoutTreeNode<()> for ChoiceWidget {
//!     // elided for brevity
//! #   fn current_node(&self) -> &LayoutNode { &self.node }
//! #   fn child_count(&self) -> usize { 2 }
//! #   fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'_>> {
//! #       match ndx {
//! #           0 => Some(self.choice1.lnode()),
//! #           1 => Some(self.choice2.lnode()),
//! #           _ => None
//! #       }
//! #   }
//! }
//! impl Widget for ChoiceWidget {
//!     fn lnode(&self) -> LayoutNodeAccess<'_> {
//!         LayoutNodeAccess::new::<()>(self)
//!     }
//! }
//! ```

use std::{ops::Deref, rc::Rc};

use cache::{Layer, NodeState};
use kahlo::math::{PixelBox, PixelSideOffsets};

mod arr;
mod cache;
mod calc;
mod dump;

pub use cache::{LayoutCache, LayoutNodeID};
pub use calc::recalculate;

pub use dump::*;

#[doc(hidden)]
pub struct CellMarker;
/// Type representing a single cell's coordinates in a table.
pub type TableCell = euclid::Point2D<usize, CellMarker>;
/// Type representing the total size of a table.
pub type TableSize = euclid::Size2D<usize, CellMarker>;

/// Sizing policy for a layout dimension. Defaults to no minimum or desired size and a low-weighted
/// desire to take up slack space.
#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
pub struct SizePolicy {
    /// The minimum number of pixels required along this dimension to avoid overlap and other
    /// issues.
    pub minimum: usize,
    /// The number of pixels requested to have a comfortable amount of space and avoid ugly-looking
    /// UI.
    pub desired: usize,
    /// How much to take up any remaining slack space.
    pub slack_weight: usize,
}

impl Default for SizePolicy {
    fn default() -> Self {
        Self {
            minimum: 0,
            desired: 0,
            slack_weight: 1,
        }
    }
}

impl SizePolicy {
    pub fn zero() -> Self {
        Self {
            minimum: 0,
            desired: 0,
            slack_weight: 0,
        }
    }

    pub fn expanding(weight: usize) -> Self {
        Self {
            minimum: 0,
            desired: 0,
            slack_weight: weight,
        }
    }

    pub fn fixed(size: usize) -> Self {
        Self {
            minimum: size,
            desired: size,
            slack_weight: 0,
        }
    }

    pub fn with_minimum(mut self, new_min: usize) -> Self {
        self.minimum = new_min;
        self.desired = self.desired.max(new_min);
        self
    }
    pub fn with_desired(mut self, new_desire: usize) -> Self {
        self.desired = new_desire;
        self
    }
    pub fn with_slack(mut self, new_slack: usize) -> Self {
        self.slack_weight = new_slack;
        self
    }

    pub fn max(self, rhs: SizePolicy) -> SizePolicy {
        Self {
            minimum: self.minimum.max(rhs.minimum),
            desired: self.desired.max(rhs.desired),
            slack_weight: self.slack_weight.max(rhs.slack_weight),
        }
    }

    pub fn max_preserve_slack(self, rhs: SizePolicy) -> SizePolicy {
        Self {
            minimum: self.minimum.max(rhs.minimum),
            desired: self.desired.max(rhs.desired),
            slack_weight: self.slack_weight,
        }
    }
}

#[doc(hidden)]
pub struct SizePolicyTag;
/// A 2D size policy type using [`euclid`].
pub type SizePolicy2D = euclid::Size2D<SizePolicy, SizePolicyTag>;

impl std::ops::Add<Self> for SizePolicy {
    type Output = Self;
    fn add(self, rhs: Self) -> Self::Output {
        Self {
            minimum: self.minimum + rhs.minimum,
            desired: self.desired + rhs.desired,
            slack_weight: self.slack_weight + rhs.slack_weight,
        }
    }
}

/// How to horizontally align a smaller node inside a larger node.
#[derive(Clone, Copy, PartialEq, Debug)]
pub enum HorizontalAlignment {
    Left,
    Centre,
    Right,
}

impl Default for HorizontalAlignment {
    fn default() -> Self {
        Self::Centre
    }
}

/// How to vertically align a smaller node inside a larger node.
#[derive(Clone, Copy, PartialEq, Debug)]
pub enum VerticalAlignment {
    Top,
    Centre,
    Bottom,
}

impl Default for VerticalAlignment {
    fn default() -> Self {
        Self::Centre
    }
}

/// What sort of layout does a node represent?
#[derive(Clone, Copy, PartialEq, Debug)]
pub enum NodeBehaviour {
    /// A fixed rendering area, probably a root node.
    Fixed { area: PixelBox },
    /// An ordinary box sitting inside another node, hinting its size and receiving a final
    /// size assignment from its parent.
    Element,
    /// A node that floats above other content, anchored with a zero-size flexbox.
    Anchored,
}

/// Represents a layout style for a layout node to arrange its children with.
#[derive(Clone)]
pub enum ChildArrangement {
    Custom(std::rc::Rc<dyn arr::ArrangementCalculator>),
    /// Node that wraps a single other node
    Container,
    /// Node that arranges its children in a vertical column
    Column,
    /// Node that arranges its children in a horizontal row
    Row,
    /// Node that arranges its children in a 2D table using [`TableCell`]
    Table,
}

impl std::fmt::Debug for ChildArrangement {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Custom(_) => f.write_str("ChildArrangement::Custom"),
            Self::Container => f.write_str("ChildArrangement::Container"),
            Self::Column => f.write_str("ChildArrangement::Column"),
            Self::Row => f.write_str("ChildArrangement::Row"),
            Self::Table => f.write_str("ChildArrangement::Table"),
        }
    }
}

impl Deref for ChildArrangement {
    type Target = dyn arr::ArrangementCalculator;
    fn deref(&self) -> &Self::Target {
        match self {
            Self::Custom(calc) => calc.as_ref(),
            Self::Container => &arr::ContainerArrangement,
            Self::Column => &arr::LineArrangement::Column,
            Self::Row => &arr::LineArrangement::Row,
            Self::Table => &arr::TableArrangement,
        }
    }
}

/// Type representing a single node in the layout tree.
///
/// Notably, this struct is unaware of any edges it may participate in; see the module
/// documentation for more information.
pub struct LayoutNode {
    /// Human-readable label for making layout tree dumps easier to read.
    label: Option<String>,
    /// Unique identifier for this LayoutNode
    cache_key: LayoutNodeID,
    /// Reference to the global layout item cache.
    cache: Rc<LayoutCache>,
    /// Layout behaviour, or how this node behaves with respect to its parent.
    behaviour: NodeBehaviour,
    /// Child arrangement, or how this node arranges its children.
    child_arrangement: ChildArrangement,
    /// Width policy: how does this widget take up horizontal space?
    width_policy: SizePolicy,
    /// Height policy: how does this widget take up vertical space?
    height_policy: SizePolicy,
    /// Horizontal alignment of this node inside its parent.
    halign: HorizontalAlignment,
    /// Vertical alignment of this node inside its parent.
    valign: VerticalAlignment,
    /// User-exposed margins, or spacing between the parent and the render area of this node.
    margin: PixelSideOffsets,
    /// Table row/column assignment for this node inside its parent.
    table_cell: Option<TableCell>,
}

impl std::fmt::Debug for LayoutNode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("LayoutNode")
            .field("label", &self.label)
            .field("cache_key", &self.cache_key)
            .field("behaviour", &self.behaviour)
            .field("child_arrangement", &self.child_arrangement)
            .field("width_policy", &self.width_policy)
            .field("height_policy", &self.height_policy)
            .field("margin", &self.margin)
            .finish()
    }
}

impl LayoutNode {
    pub fn new(cache: Rc<LayoutCache>) -> Self {
        let cache_key = LayoutNodeID::generate();
        cache.update_queue.borrow_mut().push(cache_key);
        Self {
            label: None,
            cache_key,
            cache,
            behaviour: NodeBehaviour::Element,
            child_arrangement: ChildArrangement::Column,
            width_policy: SizePolicy::default(),
            height_policy: SizePolicy::default(),
            halign: HorizontalAlignment::default(),
            valign: VerticalAlignment::default(),
            margin: PixelSideOffsets::new_all_same(0),
            table_cell: None,
        }
    }

    /// Get the [`LayoutNodeID`] for this node.
    pub fn id(&self) -> LayoutNodeID {
        self.cache_key
    }

    /// Acquire a reference to the [`LayoutCache`] this node belongs to.
    pub fn cache(&self) -> &Rc<LayoutCache> {
        &self.cache
    }

    fn with_state<R, F: FnOnce(&mut NodeState) -> R>(&self, f: F) -> Option<R> {
        self.cache.with_state(self.cache_key, f)
    }

    /// Mark as requiring a relayout.
    pub fn relayout(&self) {
        self.cache.update_queue.borrow_mut().push(self.cache_key);
    }

    pub fn relayout_tree(&self) {
        let mut to_mark: Vec<LayoutNodeID> = vec![self.cache_key];
        while let Some(next) = to_mark.pop() {
            self.cache.with_state(next, |ns| {
                ns.needs_update = true;
                to_mark.extend(ns.children.iter());
            });
        }
    }

    pub fn render_area(&self) -> Option<PixelBox> {
        self.with_state(|ns| ns.area).flatten()
    }

    /// Checks if node needs to be rerendered, clearing the flag if so.
    pub fn render_check(&self) -> bool {
        self.with_state(|ns| {
            let ret = ns.needs_render;
            ns.needs_render = false;
            ret
        })
        .unwrap_or(false)
    }

    /// Mark as requiring a rerender
    pub fn render_needed(&self) {
        self.cache.render_queue.borrow_mut().push(self.cache_key);
    }

    pub fn make_new_node(&self) -> Self {
        Self::new(self.cache.clone())
    }
}

/// Accessors
impl LayoutNode {
    pub fn label(&self) -> Option<&str> {
        self.label.as_deref()
    }
    pub fn set_label(&mut self, to: impl AsRef<str>) {
        self.label = Some(to.as_ref().to_string());
    }

    pub fn behaviour(&self) -> NodeBehaviour {
        self.behaviour
    }
    pub fn set_behaviour(&mut self, mode: NodeBehaviour) -> &mut Self {
        if self.behaviour != mode {
            self.behaviour = mode;
            self.relayout();
        }
        self
    }

    pub fn arrangement(&self) -> &ChildArrangement {
        &self.child_arrangement
    }
    pub fn set_arrangement(&mut self, arr: ChildArrangement) -> &mut Self {
        self.child_arrangement = arr;
        self.relayout();
        self
    }

    pub fn width_policy(&self) -> SizePolicy {
        self.width_policy
    }
    pub fn set_width_policy(&mut self, policy: SizePolicy) -> &mut Self {
        self.width_policy = policy;
        self.relayout();
        self
    }
    pub fn height_policy(&self) -> SizePolicy {
        self.height_policy
    }
    pub fn set_height_policy(&mut self, policy: SizePolicy) -> &mut Self {
        self.height_policy = policy;
        self.relayout();
        self
    }

    pub fn halign(&self) -> HorizontalAlignment {
        self.halign
    }
    pub fn set_halign(&mut self, halign: HorizontalAlignment) -> &mut Self {
        self.halign = halign;
        self.relayout();
        self
    }
    pub fn valign(&self) -> VerticalAlignment {
        self.valign
    }
    pub fn set_valign(&mut self, valign: VerticalAlignment) -> &mut Self {
        self.valign = valign;
        self.relayout();
        self
    }

    pub fn margin(&self) -> PixelSideOffsets {
        self.margin
    }
    pub fn margin_mut(&mut self) -> &mut PixelSideOffsets {
        self.relayout();
        &mut self.margin
    }
    pub fn margin_as_policy(&self) -> SizePolicy2D {
        SizePolicy2D::new(
            SizePolicy::fixed((self.margin.left + self.margin.right) as usize),
            SizePolicy::fixed((self.margin.top + self.margin.bottom) as usize),
        )
    }

    pub fn table_cell(&self) -> &Option<TableCell> {
        &self.table_cell
    }
    pub fn set_table_cell(&mut self, cell: Option<TableCell>) {
        self.relayout();
        self.table_cell = cell;
    }
}

impl Clone for LayoutNode {
    fn clone(&self) -> Self {
        Self {
            label: self.label.clone(),
            cache_key: LayoutNodeID::generate(),
            cache: self.cache.clone(),
            behaviour: self.behaviour,
            child_arrangement: self.child_arrangement.clone(),
            width_policy: self.width_policy,
            height_policy: self.height_policy,
            halign: self.halign,
            valign: self.valign,
            margin: self.margin,
            table_cell: self.table_cell,
        }
    }
}

impl Drop for LayoutNode {
    fn drop(&mut self) {
        self.cache.update_queue.borrow_mut().push(self.cache_key);
    }
}

/// Iterator implementation to iterate across children of a LayoutNodeContainer
#[derive(Clone)]
struct LayoutChildIter<'l> {
    lna: &'l LayoutNodeAccess<'l>,
    next_index: usize,
}

impl<'l> LayoutChildIter<'l> {
    fn new(lna: &'l LayoutNodeAccess) -> Self {
        Self { lna, next_index: 0 }
    }
}

impl<'l> Iterator for LayoutChildIter<'l> {
    type Item = LayoutNodeAccess<'l>;

    fn next(&mut self) -> Option<Self::Item> {
        let index = self.next_index;
        if index >= self.lna.child_count() {
            None
        } else {
            self.next_index += 1;
            self.lna.child(index)
        }
    }
}

/// Accessor trait for a struct that stores edge information for the layout tree.
///
/// This trait accepts a tag type parameter to allow a single struct to store multiple layout nodes
/// with internal edges. Use a `Tag` type of `()` or similar to denote a struct that only
/// stores a single layout node.
pub trait LayoutTreeNode<Tag: 'static = ()> {
    fn current_node(&self) -> &LayoutNode;
    fn child_count(&self) -> usize;
    fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'_>>;
}

const DYNPTR_WIDTH: usize =
    std::mem::size_of::<&'static dyn LayoutTreeNode<()>>() / std::mem::size_of::<*const ()>();
type DynPtr = [*const (); DYNPTR_WIDTH];

fn current_node_dispatch<'l, Tag: 'static>(dynptr: DynPtr) -> &'l LayoutNode {
    unsafe { std::mem::transmute::<_, &'l dyn LayoutTreeNode<Tag>>(dynptr).current_node() }
}

fn child_count_dispatch<'l, Tag: 'static>(dynptr: DynPtr) -> usize {
    unsafe { std::mem::transmute::<_, &'l dyn LayoutTreeNode<Tag>>(dynptr).child_count() }
}

fn child_dispatch<'l, Tag: 'static>(dynptr: DynPtr, ndx: usize) -> Option<LayoutNodeAccess<'l>> {
    unsafe { std::mem::transmute::<_, &'l dyn LayoutTreeNode<Tag>>(dynptr).child(ndx) }
}

/// Edge-aware accessor for a [`LayoutNode`].
///
/// Use this struct if you wish to accept arbitrary [`LayoutNode`]s and require information about
/// the layout tree as well.
///
/// <details><summary>Implementation details</summary>
///
/// `LayoutNodeAccess` is currently implemented with some unsafe code under the hood due to
/// limitations of the Rust type system. Specifically, this type has three goals:
/// 1. Avoid being a generic type for ease of use
/// 2. Represent [`LayoutTreeNode`] instances with arbitrary tags
/// 3. For performance reasons, avoid dynamic allocation: all storage should be interior to the type
///
/// The most obvious way to implement this type is by storing several functions as e.g. `Box<Fn() ->
/// usize>` to proxy the methods in the tagged `LayoutTreeNode` instance, but this trivially fails
/// goal #3. The current implementation stores a copy of the instance and the vtable pointer to
/// perform dynamic dispatch directly, which maintains type-safety invariants because the only
/// varying parameter is the tag type.
/// </details>

#[derive(Clone, Copy)]
pub struct LayoutNodeAccess<'l> {
    current_node_func: fn(DynPtr) -> &'l LayoutNode,
    child_count_func: fn(DynPtr) -> usize,
    child_func: fn(DynPtr, ndx: usize) -> Option<LayoutNodeAccess<'l>>,
    dynptr: DynPtr,
}

impl<'l> LayoutNodeAccess<'l> {
    pub fn new<Tag: 'static>(from: &'l dyn LayoutTreeNode<Tag>) -> Self {
        assert_eq!(
            std::mem::size_of::<&dyn LayoutTreeNode<Tag>>(),
            std::mem::size_of::<DynPtr>()
        );
        assert_eq!(
            std::mem::align_of::<&dyn LayoutTreeNode<Tag>>(),
            std::mem::align_of::<DynPtr>()
        );
        Self {
            current_node_func: current_node_dispatch::<Tag>,
            child_count_func: child_count_dispatch::<Tag>,
            child_func: child_dispatch::<Tag>,
            dynptr: unsafe { std::mem::transmute(from) },
        }
    }

    pub fn current_node(&self) -> &'l LayoutNode {
        (self.current_node_func)(self.dynptr)
    }

    pub fn child_count(&self) -> usize {
        (self.child_count_func)(self.dynptr)
    }

    pub fn child(&self, ndx: usize) -> Option<LayoutNodeAccess<'l>> {
        (self.child_func)(self.dynptr, ndx)
    }

    pub fn child_iter(&self) -> impl Iterator<Item = LayoutNodeAccess> + Clone {
        LayoutChildIter::new(&self)
    }
}

impl<'l> Deref for LayoutNodeAccess<'l> {
    type Target = LayoutNode;
    fn deref(&self) -> &Self::Target {
        (self.current_node_func)(self.dynptr)
    }
}

/// Helper struct to store a leaf LayoutNode and automatically provide a [`LayoutTreeNode`]
/// implementation.
///
/// This helps avoid repetitive reimplementations of [`LayoutTreeNode`] when a container stores a
/// node with no children.
///
/// If you're reading this documentation, you probably want the [`LeafLayoutNode::access`]
/// function, which gives a [`LayoutNodeAccess`].
pub struct LeafLayoutNode(LayoutNode);

impl LeafLayoutNode {
    pub fn new(ln: LayoutNode) -> Self {
        Self(ln)
    }

    pub fn access(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(self)
    }
}

impl From<LayoutNode> for LeafLayoutNode {
    fn from(value: LayoutNode) -> Self {
        Self::new(value)
    }
}

impl std::ops::Deref for LeafLayoutNode {
    type Target = LayoutNode;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl std::ops::DerefMut for LeafLayoutNode {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl LayoutTreeNode<()> for LeafLayoutNode {
    fn current_node(&self) -> &LayoutNode {
        &self.0
    }
    fn child(&self, _ndx: usize) -> Option<LayoutNodeAccess<'_>> {
        None
    }
    fn child_count(&self) -> usize {
        0
    }
}

/*macro_rules! layout_tree {
    ($ctype:ident, $($tag:ident => $nname:ident | [$($ctag:ident),*]),*) => {

    };
}*/
