use super::{do_fit, ArrangementCalculator};
use crate::{
    layout::{LayoutNodeAccess, SizePolicy, SizePolicy2D},
    math::{PixelBox, PixelPoint},
};
use std::ops::Add;

#[derive(Clone, Debug)]
pub struct TableArrangement;

#[derive(Default)]
struct TableState {
    row_policies: Vec<SizePolicy>,
    col_policies: Vec<SizePolicy>,
}

impl TableArrangement {
    fn build_table_state(&self, node: &LayoutNodeAccess) -> TableState {
        let mut tstate = TableState {
            row_policies: vec![],
            col_policies: vec![],
        };

        // pull row/column information from child node properties
        for ch in node.child_iter() {
            let Some(cell) = ch.table_cell() else {
                continue;
            };

            if cell.x >= tstate.col_policies.len() {
                tstate.col_policies.resize(cell.x + 1, SizePolicy::zero());
            }
            if cell.y >= tstate.row_policies.len() {
                tstate.row_policies.resize(cell.y + 1, SizePolicy::zero());
            }

            let child_policy = ch.with_state(|v| v.net_policy).unwrap();
            tstate.col_policies[cell.x] = tstate.col_policies[cell.x].max(child_policy.width);
            tstate.row_policies[cell.y] = tstate.row_policies[cell.y].max(child_policy.height);
        }

        tstate
    }
}

impl ArrangementCalculator for TableArrangement {
    fn arrange_step(&self, node: LayoutNodeAccess) -> SizePolicy2D {
        let tstate = self.build_table_state(&node);

        // if there are no table cells...
        if tstate.row_policies.len() == 0 || tstate.col_policies.len() == 0 {
            return SizePolicy2D::new(node.width_policy(), node.height_policy())
                + node.margin_as_policy();
        }

        let net_rows = tstate
            .row_policies
            .into_iter()
            .reduce(SizePolicy::add)
            .unwrap();
        let net_cols = tstate
            .col_policies
            .into_iter()
            .reduce(SizePolicy::add)
            .unwrap();

        SizePolicy2D::new(
            node.width_policy.max_preserve_slack(net_cols),
            node.height_policy.max_preserve_slack(net_rows),
        ) + node.margin_as_policy()
    }

    fn layout_step(&self, node: LayoutNodeAccess, inside: PixelBox) {
        let inside = super::do_align(&node, inside);

        node.with_state(|ns| {
            if Some(inside) != ns.area {
                ns.area = Some(inside);
                ns.needs_render = true;
            }
        });

        let inside = inside.inner_box(node.margin());

        let tstate = self.build_table_state(&node);

        let mut col_offsets = vec![0];
        col_offsets.extend(do_fit(inside.width(), tstate.col_policies.into_iter()));
        let mut row_offsets = vec![0];
        row_offsets.extend(do_fit(inside.height(), tstate.row_policies.into_iter()));

        let child_offset = inside.min.to_vector();

        for ch in node.child_iter() {
            let Some(cell) = ch.table_cell() else {
                continue;
            };

            let cbox = PixelBox::new(
                PixelPoint::new(col_offsets[cell.x], row_offsets[cell.y]),
                PixelPoint::new(col_offsets[cell.x + 1], row_offsets[cell.y + 1]),
            );

            ch.child_arrangement
                .layout_step(ch, cbox.translate(child_offset));
        }
    }
}
