use std::{cell::RefCell, collections::HashMap, rc::Rc};

use super::SizePolicy2D;

/// Unique per-node ID value.
///
/// This struct is a thin wrapper around a `usize`. Generally, you should never have to interact
/// with this type unless you are working on patina itself.
#[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy)]
pub struct LayoutNodeID(usize);
static NEXT_NODE_ID: std::sync::atomic::AtomicUsize = std::sync::atomic::AtomicUsize::new(1);
impl LayoutNodeID {
    pub fn generate() -> Self {
        Self(NEXT_NODE_ID.fetch_add(1, std::sync::atomic::Ordering::SeqCst))
    }
}

impl Into<usize> for LayoutNodeID {
    fn into(self) -> usize {
        self.0
    }
}

#[derive(Debug)]
pub struct NodeState {
    pub(super) needs_update: bool,
    pub(super) needs_render: bool,
    pub(super) net_policy: SizePolicy2D,
    pub(super) layer: Layer,
    pub(super) area: Option<kahlo::math::PixelBox>,
    pub(super) children: Vec<LayoutNodeID>,
}

/// This struct stores meta-information about each node, such as its calculated render area.
///
/// All nodes have a refcounted reference to an instance of this struct.
#[derive(Debug, Default)]
pub struct LayoutCache {
    pub(super) update_queue: RefCell<Vec<LayoutNodeID>>,
    pub(super) render_queue: RefCell<Vec<LayoutNodeID>>,
    states: RefCell<HashMap<LayoutNodeID, NodeState>>,
    parents: RefCell<HashMap<LayoutNodeID, LayoutNodeID>>,
}

impl LayoutCache {
    pub fn new() -> Rc<LayoutCache> {
        Rc::new(Self {
            update_queue: Default::default(),
            render_queue: Default::default(),
            states: Default::default(),
            parents: Default::default(),
        })
    }

    pub fn propagate_relayouts(&self) {
        let mut updq = self.update_queue.borrow_mut();
        let mut states = self.states.borrow_mut();
        let parents = self.parents.borrow();
        while let Some(next) = updq.pop() {
            if let Some(v) = states.get_mut(&next) {
                if !v.needs_update {
                    if let Some(p) = parents.get(&next) {
                        updq.push(*p)
                    }
                    v.needs_update = true;
                }
            }
        }
    }

    pub fn propagate_rerenders(&self) {
        let mut renq = self.render_queue.borrow_mut();
        let mut states = self.states.borrow_mut();

        while let Some(next) = renq.pop() {
            if let Some(v) = states.get_mut(&next) {
                v.needs_render = true;
                renq.extend(v.children.iter());
            }
        }
    }

    pub fn with_state<R, F: FnOnce(&mut NodeState) -> R>(
        &self,
        ckey: LayoutNodeID,
        f: F,
    ) -> Option<R> {
        self.states.borrow_mut().get_mut(&ckey).map(f)
    }

    pub fn has_state_for(&self, ckey: LayoutNodeID) -> bool {
        self.states.borrow().contains_key(&ckey)
    }

    pub fn store(&self, ckey: LayoutNodeID, parent: Option<LayoutNodeID>, ns: NodeState) {
        self.states.borrow_mut().insert(ckey, ns);
        if let Some(pck) = parent {
            self.parents.borrow_mut().insert(ckey, pck);
        }
    }
}

#[derive(Clone, Copy, PartialEq, PartialOrd, Debug, Default)]
pub struct Layer(u16, u16);

impl Layer {
    pub fn nest(&self) -> Self {
        Layer(self.0, self.1 + 1)
    }

    pub fn float(&self) -> Self {
        Layer(self.0 + 1, 0)
    }
}
