use std::{cell::RefCell, rc::Rc};

use kahlo::{
    math::{PixelBox, PixelPoint, PixelSize},
    prelude::{BitmapAccess, KahloOps},
};

use crate::layout::LayoutNode;

pub struct TextCache {
    text: String,
    font: Rc<fontdue::Font>,
    pxsize: f32,
    pub(crate) x_offsets: RefCell<Vec<u32>>,
    rendered: RefCell<Option<kahlo::Alphamap>>,
}

impl TextCache {
    pub fn new_with_font_and_size(font: Rc<fontdue::Font>, pxsize: f32) -> Self {
        Self {
            text: String::new(),
            font,
            pxsize,
            x_offsets: vec![].into(),
            rendered: Default::default(),
        }
    }
    pub fn text(&self) -> &str {
        &self.text
    }
    pub fn update(&mut self, text: &str) {
        // if there are actually no changes, ignore
        if self.text == text {
            return;
        }
        self.text.clear();
        self.text += text;
        self.rendered.take();
    }
    pub fn change<R, F: Fn(&mut String) -> R>(&mut self, func: F) -> R {
        let rval = func(&mut self.text);
        self.rendered.take();
        rval
    }

    pub fn size_hint(&self) -> PixelSize {
        self.calculate_size(self.pxsize, &self.text)
    }

    pub fn update_node_size_hint(&self, node: &mut LayoutNode) {
        let sz = self.size_hint();

        let wp = node.width_policy().with_minimum(sz.width as usize);
        node.set_width_policy(wp);
        let hp = node.height_policy().with_minimum(sz.height as usize);
        node.set_height_policy(hp);
    }
}

// internal functions
impl TextCache {
    pub(crate) fn with_rendered<R>(&self, func: impl FnOnce(&kahlo::Alphamap) -> R) -> Option<R> {
        let mut rendered = self.rendered.borrow_mut();
        if let Some(amap) = rendered.as_ref() {
            return Some(func(amap));
        }

        *rendered = self.render(self.pxsize);

        rendered.as_ref().map(func)
    }

    fn render(&self, pxsize: f32) -> Option<kahlo::Alphamap> {
        // for now assume the input text is a single line
        let amap = self.render_line(pxsize, &self.text);
        Some(amap)
    }

    fn generate_baseline_offsets<'l, 's: 'l>(
        &'s self,
        pxsize: f32,
        text: &'l str,
    ) -> impl 'l + Iterator<Item = (u32, u32, char)> {
        let mut running_offset = 0;
        let mut last_char = None;

        text.chars().map(move |ch| {
            let metrics = self.font.metrics(ch, pxsize);

            let glyph_width = if let Some(last) = last_char.take() {
                metrics.advance_width + self.font.horizontal_kern(last, ch, pxsize).unwrap_or(0.)
            } else {
                metrics.advance_width
            };

            last_char = Some(ch);

            let expose = (running_offset, glyph_width.round() as u32, ch);
            running_offset += glyph_width.round() as u32;
            expose
        })
    }

    fn calculate_size(&self, pxsize: f32, text: &str) -> PixelSize {
        // pass: determine required size
        let total_width = match self.generate_baseline_offsets(pxsize, text).last() {
            Some(ofs) => ofs.0 + ofs.1,
            None => 0,
        };
        let line_metrics = self.font.horizontal_line_metrics(pxsize).unwrap();
        let line_height = line_metrics.new_line_size.ceil() as i32;

        PixelSize::new(total_width as i32, line_height)
    }

    fn render_line(&self, pxsize: f32, text: &str) -> kahlo::Alphamap {
        // pass 1: determine required size
        let alphamap_size = self.calculate_size(pxsize, text);

        let line_metrics = self.font.horizontal_line_metrics(pxsize).unwrap();
        let baseline = line_metrics.ascent.ceil() as usize;
        let mut alphamap =
            kahlo::Alphamap::new(alphamap_size.width as usize, alphamap_size.height as usize);

        let mut x_offsets = self.x_offsets.borrow_mut();
        x_offsets.clear();

        // pass: generate alphamap
        for (offset, _width, ch) in self.generate_baseline_offsets(pxsize, text) {
            let (metrics, raster) = self.font.rasterize(ch, pxsize);
            let character_alphamap = kahlo::BitmapRef::<kahlo::formats::A8>::new(
                raster.as_slice(),
                metrics.width,
                metrics.height,
            );

            x_offsets.push(offset);

            alphamap.copy_from(
                &character_alphamap,
                PixelBox::from_size(character_alphamap.size()),
                PixelPoint::new(
                    metrics.xmin + offset as i32,
                    baseline as i32 - (metrics.height as i32 + metrics.ymin),
                ),
            );
        }

        alphamap
    }
}
