use crate::{
    input::MouseButton,
    layout::{
        HorizontalAlignment, LayoutNode, LayoutNodeAccess, LayoutTreeNode, LeafLayoutNode,
        SizePolicy,
    },
    prelude::*,
    render::{ColourChoice, RenderTarget, TextCache},
    ui::UIHandle,
};
use kahlo::math::{PixelBox, PixelPoint, PixelSideOffsets, PixelSize};
use winit::keyboard::{Key, NamedKey};

pub enum CursorPosition {
    Start,
    Index(usize),
    End,
}

/*impl CursorPosition {

}*/

pub struct TextEdit<C: Component> {
    lnode: LeafLayoutNode,
    text: TextCache,
    cursor_pos: CursorPosition,
    onchange: Option<Box<dyn Fn(&str) -> Option<C::Msg>>>,
    focused: bool,
}

impl<C: Component> TextEdit<C> {
    pub fn new(uih: &UIHandle) -> Self {
        let minheight = (uih.theme().ui_font_size * 1.5).round() as usize;
        let mut lnode = uih.new_layout_node();
        lnode
            .set_width_policy(SizePolicy::expanding(1))
            .set_height_policy(SizePolicy::fixed(minheight));

        Self {
            lnode: LeafLayoutNode::new(lnode),
            text: TextCache::new_with_font_and_size(
                uih.theme().ui_font.clone(),
                uih.theme().ui_font_size,
            ),
            cursor_pos: CursorPosition::Start,
            onchange: None,
            focused: false,
        }
    }

    pub fn set_text(&mut self, to: &str) {
        self.text.update(to);
    }

    /*pub fn set_cursor_position(&mut self, to: CursorPosition) {

    }*/
}

impl<C: Component> Widget<C> for TextEdit<C> {
    fn poll(
        &mut self,
        _uih: &mut UIHandle,
        input_state: Option<&crate::input::InputState>,
    ) -> Vec<<C as Component>::Msg> {
        let Some(istate) = input_state else {
            return vec![];
        };

        if istate.is_focused(self.lnode.id()) {
            let mut key_press_eaten = true;
            if let Some(kp) = &istate.keypress {
                match kp {
                    Key::Named(NamedKey::Backspace) => {
                        self.text.change(String::pop);
                        self.lnode.render_needed();
                    }
                    Key::Named(NamedKey::Delete) => {
                        self.text.change(String::pop);
                        self.lnode.render_needed();
                    }
                    Key::Named(NamedKey::Space) => {
                        self.text.change(|v| v.push(' '));
                        self.lnode.render_needed();
                    }
                    Key::Named(NamedKey::Enter) => {
                        istate.clear_focus();
                    }
                    Key::Named(_) => {}
                    _ => key_press_eaten = false,
                }
            }
            if !key_press_eaten {
                if let Some(text) = &istate.text_input {
                    println!("adding text {text:?}");
                    self.text.change(|s| s.push_str(text.as_str()));
                    self.lnode.render_needed();
                }
            }
        }

        let render_area = self.lnode.render_area().unwrap();
        if istate
            .mouse_status_in(render_area, MouseButton::Left)
            .is_click()
        {
            istate.focus_on(self.lnode.id());
            println!("click offset: {:?}", istate.mouse.pos - render_area.min);
        }

        let was_focused = self.focused;
        self.focused = istate.is_focused(self.lnode.id());

        if self.focused != was_focused {
            println!("self.focused is now {}", self.focused);
            self.lnode.render_needed();
        }

        vec![]
    }

    fn layout_node(&self) -> LayoutNodeAccess {
        self.lnode.access()
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.lnode
    }

    fn render<'r, 'data: 'r>(&self, target: &mut RenderTarget<'r, 'data>) {
        target
            .with_node(&self.lnode)
            .fill(ColourChoice::Background)
            .simple_text(&self.text, ColourChoice::Foreground)
            .seq(|target| {
                if self.focused {
                    let cbar_offset = match self.cursor_pos {
                        CursorPosition::Start => 0,
                        CursorPosition::Index(idx) => {
                            *self.text.x_offsets.borrow().get(idx).unwrap() as i32
                        }
                        CursorPosition::End => self.text.size_hint().width,
                    };
                    target.fill_relative_box(
                        PixelBox::from_origin_and_size(
                            PixelPoint::new(cbar_offset, 0),
                            PixelSize::new(2, self.text.size_hint().height),
                        ),
                        ColourChoice::Foreground,
                    );
                }
            });
    }
}
